# Copyright (c) 2006-2007 The Regents of The University of Michigan
# Copyright (c) 2009 Advanced Micro Devices, Inc.
# Copyright (c) 2011 Mark D. Hill and David A. Wood
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met: redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer;
# redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution;
# neither the name of the copyright holders nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: Brad Beckmann

import m5
from m5.objects import *
from m5.defines import buildEnv
from m5.util import panic
import math

#
# Note: the L1 Cache latency is only used by the sequencer on fast path hits
#
class L1Cache(RubyCache):
	latency = 14

#
# Note: the L2 Cache latency is not currently used
#
class L2Cache(RubyCache):
	latency = 10

def create_system(options, system, cpu_sequencers, all_cntrls, ruby_system):

	if not buildEnv['GPGPU_SIM']:
		panic("This script requires the GPGPU-sim to be built.")

	#
	# The ruby network creation expects the list of nodes in the system to be
	# consistent with the NetDest list.  Therefore the l1 controller nodes must be
	# listed before the directory nodes and directory nodes before dma nodes, etc.
	#
	l1_cntrl_nodes = []
	block_size_bits = int(math.log(options.cacheline_size, 2))


	#cache for stream processor
	#
	for i in xrange(options.num_sc):
		#
		# First create the Ruby objects associated with this cpu
		#

		#shader should never access the instruction cache
		l1i_cache = L1Cache(size = options.l1i_size,
							assoc = options.l1i_assoc,
							start_index_bit = block_size_bits)
		l1d_cache = L1Cache(size = options.sc_l1_size,
							assoc = options.sc_l1_assoc,
							start_index_bit = block_size_bits, replacement_policy = "LRU")
		l2_cache = L2Cache(size = options.sc_l2_size,
							assoc = options.l2_assoc,
							start_index_bit = block_size_bits, replacement_policy = "LRU")

		cpu_seq = RubySequencer(version = options.num_cpus+i,
								icache = l1i_cache,
								dcache = l1d_cache,
								physMemPort = system.physmem.port,
								physmem = system.physmem,
								access_phys_mem = False,
                               ruby_system = ruby_system)

		#if piobus != None:
		#	cpu_seq.pio_port = piobus.port

		l1_cntrl = L1Cache_Controller(version = options.num_cpus+i,
                                                                                cntrl_id = len(all_cntrls)+i,
										sequencer = cpu_seq,
										L1IcacheMemory = l1i_cache,
										L1DcacheMemory = l1d_cache,
										L2cacheMemory = l2_cache,
										no_mig_atomic = not \
										options.allow_atomic_migration,
                               ruby_system = ruby_system)

		if options.recycle_latency:
			l1_cntrl.recycle_latency = options.recycle_latency

		if (i<10):
			exec("system.l1_cntrl_sc0%d = l1_cntrl" % i)
		else:
			exec("system.l1_cntrl_sc%d = l1_cntrl" % i)
		#
		# Add controllers and sequencers to the appropriate lists
		#
		cpu_sequencers.append(cpu_seq)
		l1_cntrl_nodes.append(l1_cntrl)

	l1i_cache = L1Cache(size = "4096B",
						assoc = 2,
						start_index_bit = block_size_bits)
	l1d_cache = L1Cache(size = "4096B",
						assoc = 2,
						start_index_bit = block_size_bits)
	l2_cache = L2Cache(size = "4096B",
						assoc = 2,
						start_index_bit = block_size_bits)

	#
	# Only one unified L1 cache exists.  Can cache instructions and data.
	#
	cpu_seq = RubySequencer(version = options.num_cpus+options.num_sc,
							   icache = l1i_cache,
							   dcache = l1d_cache,
							   physMemPort = system.physmem.port,
							   physmem = system.physmem,
							   access_phys_mem = True,
                               ruby_system = ruby_system)

	l1_cntrl = L1Cache_Controller(version = options.num_cpus+options.num_sc,
										sequencer = cpu_seq,
                                                                                cntrl_id = len(all_cntrls)+options.num_sc,
										L1IcacheMemory = l1i_cache,
										L1DcacheMemory = l1d_cache,
										L2cacheMemory = l2_cache,
										no_mig_atomic = not \
										options.allow_atomic_migration,
                               ruby_system = ruby_system)

	system.l1_cntrl_ce = l1_cntrl

	cpu_sequencers.append(cpu_seq)
	l1_cntrl_nodes.append(l1_cntrl)

	all_cntrls += l1_cntrl_nodes

	return
