#!/usr/bin/env python
# Copyright (c) 2011 Mark D. Hill and David A. Wood
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met: redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer;
# redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution;
# neither the name of the copyright holders nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import sys
import os
import getopt

defaultConfig = "configs/example/ruby_gpgpusim.py"
defaultDP = "CUDA"

benchmarks = {
    "template":{"args":'', "desc":"Template from CUDA SDK", "executable":"gpgpu_ptx_sim__template"},
    "AES":{"args":"e 128 ./data/output.bmp ./data/key128.txt",
        "desc":'AES Encryption', "executable":"gpgpu_ptx_sim__aescuda"},
    "BFS":{"args":'./data/graph4096.txt', "desc":"Breadth-first search", "executable":"gpgpu_ptx_sim__bfs"},
    "BlackScholes":{"args":'', "desc":"Black-Scholes algorithm from CUDA SDK", "executable":"gpgpu_ptx_sim__BlackScholes"},
    "NN":{"args":'8', "desc":"Neural Nets", "executable":"gpgpu_ptx_sim__neuralnet"},
    "r_bfs":{"args":'./data/graph4096.txt', "desc":"Breadth-first search from Rodinia", "executable":"gpgpu_ptx_sim__bfs"},
    "ub1":{"args":'input.txt', "desc":"Dumb microbench", "executable":"gpgpu_ptx_sim__ub1"},
    "atomic_test":{"args":'', "desc":"Tests atomic operations", "executable":"gpgpu_ptx_sim__atomic"},
    "NQU":{"args":'', "desc":"Solves N Queens problem", "executable":"gpgpu_ptx_sim__nqueen"},
    "barrier_test":{"args":'', "desc":"Tests barriers", "executable":"gpgpu_ptx_sim__barrier"},
    "throughput":{"args":'524288 1 1024', "desc":"Tests throughput", "executable":"gpgpu_ptx_sim__throughput"},
    "global":{"args":'2048 1', "desc":"Tests memory latency", "executable":"gpgpu_ptx_sim__global"},
    "memcpy":{"args":'', "desc":"Tests memory copy times", "executable":"gpgpu_ptx_sim__memcpy"},
    "p_blackscholes":{"args":'1 inputs/in_16.txt result.txt', "desc":"BlackScholes from Parsec", "executable":"gpgpu_ptx_sim__blackscholes"},
    "fluidanimate":{"args":'16384 1 in_5K.fluid', "desc":"Fluidanimate from Parsec", "executable":"gpgpu_ptx_sim__fluidanimate"},
    "nw/host_sync":{"args":'128 10', "desc":"Needleman-Wunsch algorithm (synchronizes through host)", "executable":"gpgpu_ptx_sim__needle"},
    "nw/dev_sync":{"args":'128 10', "desc":"Needleman-Wunsch algorithm (synchronizes on device)", "executable":"gpgpu_ptx_sim__needle"},
}

def help():
    print "Usage: gpgpu-run.py <options> path-to-gem5-executable benchmark"
    print
    print "\t-c\t\tpath to config file, assumed to be ruby_gpgpusim.py"
    print "\t-d\t\tdesign point. Either CUDA or CUDAcc_dp4, default is CUDA"
    print "\t-f\t\tdebug flags to use"
    print "\t-F\t\tuse Fermi configuration"
    print "\t-g\t\tsimulate only the GPU, use the atomic CPU"
    print "\t-t\t\tuse gpgpu-sim memory system (default is ruby)"
    print "\t-o <args>\targuments to pass the benchmark, overrides default"
    print "\t-h, --help\tprint this help"
    print
    print "\tPossible benchmarks"
    keys = benchmarks.keys()
    keys.sort()
    for key in keys:
        if len(key) < 8:
            tab = "\t\t"
        else:
            tab = "\t"
        print '\t\t'+key+tab+benchmarks[key]["desc"]


def parseArgs(args):
    try:
        options, args = getopt.getopt(args[1:], "hgtFc:d:f:o:", ["help","fermi"])
    except getopt.GetoptError, err:
        print str(err)
        help()
        sys.exit(2)

    config = defaultConfig
    dp = defaultDP
    flags = ''
    gpuOnly = False
    fermi = False
    rubyMem = True
    benchArgs = None
    for o,a in options:
        if o in ["-h", "--help"]:
            help()
            sys.exit()
        if o in ["--fermi", "-F"]:
            fermi = True
        elif o == "-c":
            config = a
        elif o == "-d":
            dp = a
        elif o == "-f":
            flags = a
        elif o == "-g":
            gpuOnly = True
        elif o == "-t":
            rubyMem = False
        elif o == "-o":
            benchArgs = a

    if len(args) != 2:
        print "No benchmark or no exec or bad options"
        help()
        sys.exit(3)

    exe = args[0]
    bench = args[1]

    benchPath = "tests/gpgpu-sim/benchmarks/"+dp+'/'+bench+'/'

    return exe, benchPath, bench, config, flags, gpuOnly, fermi, rubyMem, benchArgs

if __name__=="__main__":
    exe, benchPath, bench, config, flags, gpuOnly, fermi, rubyMem, benchArgs = parseArgs(sys.argv)

    try:
        os.chdir(benchPath)

        benchExe = benchmarks[bench]["executable"]

        gem5Root = os.environ['GEM5_ROOT']
        exePath = gem5Root+exe
        config = gem5Root+config

        if flags:
            flags = "--debug-flags="+flags

        if fermi:
            fermi = "--fermi=True"
        else:
            fermi = ""

        if gpuOnly:
            gpuOnly = "--gpu_only=True"
        else:
            gpuOnly = ""

        if rubyMem:
            rubyMem = ""
        else:
            rubyMem = "--gpu_ruby=False"

        if not benchArgs:
            benchArgs = benchmarks[bench]["args"]

        command = exePath+" "+flags+" "+config+" "+fermi+" "+gpuOnly+" "+rubyMem+" "+" -c "+benchExe+" -o "+'"'+benchArgs+'"'
    except Exception, e:
        print "Failed to run for some reason "+str(e)
        help()
        sys.exit(100)

    print "building..."
    if os.system("make > /dev/null"):
        print "compile failed"
        sys.exit(2)
    print "DONE!"

    try:
        os.system(command)

        os.chdir("../../../../../")
    except Exception, e:
        print "Failed to run for some reason "+str(e)
        help()
        sys.exit(100)
